# nozzle.R - Report library for high-throughput data analysis pipelines.
#
# Author: Nils Gehlenborg / Harvard Medical School & Broad Institute
#
# Copyright (c) 2011-2012.

# TODO: add default state (open/closed) for sections (per default all will be closed in custom reports)

# --- global variables ----------------------------------------------------------------------------

.nozzleEnv <- new.env();

.PACKAGE.VERSION <- "0.3-1";

.ELEMENT.REPORT <- "_report_";
.ELEMENT.SECTION <- "_section_";
.ELEMENT.SUBSECTION <- "_subsection_";
.ELEMENT.SUBSUBSECTION <- "_subsubsection_";
.ELEMENT.PARAGRAPH <- "_paragraph_";
.ELEMENT.LIST <- "_list_";
.ELEMENT.PARAMETERLIST <- "_list_";
.ELEMENT.TABLE <- "_table_";
.ELEMENT.FIGURE <- "_figure_";
.ELEMENT.CITATION <- "_citation_";
.ELEMENT.RESULT <- "_result_";

.SEVERITY.LOG <- 0;
.SEVERITY.WARNING <- 1;
.SEVERITY.ERROR <- 2;

# names for popout windows
.POPOUT.FIGURE.WINDOW <- "nozzle_popout";
.POPOUT.TABLE.WINDOW <- "nozzle_popout";

DEFAULT.REPORT.FILENAME <- "nozzle";
DEFAULT.SIGNIFICANT.ENTITY <- "statistically significant findings";


# location relative to inst diretory of package
.NOZZLE.JAVASCRIPT.PATH <- "js";
.NOZZLE.JAVASCRIPT.FILE <- "nozzle.js";
.NOZZLE.JAVASCRIPT.MIN.FILE <- "nozzle.min.js";

.NOZZLE.JQUERY.MIN.FILE <- "jquery-1.5.1.min.js";
.NOZZLE.TABLESORTER.MIN.FILE <- "jquery.tablesorter.min.js";

.NOZZLE.GOOGLEANALYTICS.FILE <- "google-analytics.js";
.NOZZLE.GOOGLEANALYTICS.VARIABLE <- "GOOGLE_ANALYTICS_ID";

.NOZZLE.CSS.PATH <- "css";
.NOZZLE.CSS.FILE <- "nozzle.css";
.NOZZLE.CSS.MIN.FILE <- "nozzle.min.css";
.NOZZLE.CSS.PRINT.FILE <- "nozzle.print.css";
.NOZZLE.CSS.PRINT.MIN.FILE <- "nozzle.print.min.css";

.NOZZLE.CREDITS <- "Made with Nozzle";
.NOZZLE.URL <- "http://gdac.broadinstitute.org/nozzle";


# formatting related defaults
TABLE.SIGNIFICANT.DIGITS <- 2;


# the ordering is important. It must be public < tcga < private.
PROTECTION.PUBLIC <- 0;
PROTECTION.TCGA <- 5;
PROTECTION.GROUP <- 5;
PROTECTION.PRIVATE <- 10;

LOGO.TOP.LEFT <- 1;
LOGO.TOP.CENTER <- 2;
LOGO.TOP.RIGHT <- 4;
LOGO.BOTTOM.LEFT <- 8;
LOGO.BOTTOM.CENTER <- 16;
LOGO.BOTTOM.RIGHT <- 32;

IMAGE.TYPE.RASTER <- 0;
IMAGE.TYPE.SVG <- 1;
IMAGE.TYPE.PDF <- 2;

SECTION.CLASS.RESULTS <- "results";
SECTION.CLASS.META <- "meta";


# output types
HTML.REPORT <- "html_standalone";
HTML.FRAGMENT <- "html_fragment";
RDATA.REPORT <- "rdata";


.REFERENCE.STRING <- "##REF##";

.onLoad <- function( libname, pkgname )
{
	.initEnvCounter( ".FIGURE.COUNTER" );
	.initEnvCounter( ".TABLE.COUNTER" );
	.initEnvCounter( ".FIGURE.SUPPLEMENTARY.COUNTER" );
	.initEnvCounter( ".TABLE.SUPPLEMENTARY.COUNTER" );
	.initEnvCounter( ".CITATION.COUNTER" );
	.initEnvCounter( ".RESULT.COUNTER" ); # counter for non-empty results (assigned upon result writing)

	.initEnvCounter( ".ELEMENT.ID" );
	.initEnvCounter( ".RESULT.ID" ); # ids also for empty results (assigned upon result creation)
	
	.setGlobalReportId( "" );
	
	# TODO: check if files exist
	assign( "cssCode", readLines( file.path( system.file( .NOZZLE.CSS.PATH, package=pkgname, lib.loc=libname ), .NOZZLE.CSS.FILE), warn=FALSE ), env=.nozzleEnv );
	assign( "cssPrintCode", readLines( file.path( system.file( .NOZZLE.CSS.PATH, package=pkgname, lib.loc=libname ), .NOZZLE.CSS.PRINT.FILE), warn=FALSE ), env=.nozzleEnv );
	assign( "javaScriptCode", readLines( file.path( system.file( .NOZZLE.JAVASCRIPT.PATH, package=pkgname, lib.loc=libname ), .NOZZLE.JAVASCRIPT.FILE ), warn=FALSE ), env=.nozzleEnv );	
	assign( "cssMinCode", readLines( file.path( system.file( .NOZZLE.CSS.PATH, package=pkgname, lib.loc=libname ), .NOZZLE.CSS.MIN.FILE ), warn=FALSE ), env=.nozzleEnv );
	assign( "cssPrintMinCode", readLines( file.path( system.file( .NOZZLE.CSS.PATH, package=pkgname, lib.loc=libname ), .NOZZLE.CSS.PRINT.MIN.FILE ), warn=FALSE ), env=.nozzleEnv );
	assign( "javaScriptMinCode", readLines( file.path( system.file( .NOZZLE.JAVASCRIPT.PATH, package=pkgname, lib.loc=libname ), .NOZZLE.JAVASCRIPT.MIN.FILE ), warn=FALSE ), env=.nozzleEnv );	

	assign( "jQueryMinCode", readLines( file.path( system.file( .NOZZLE.JAVASCRIPT.PATH, package=pkgname, lib.loc=libname ), .NOZZLE.JQUERY.MIN.FILE ), warn=FALSE ), env=.nozzleEnv );	
	assign( "jQueryTableSorterMinCode", readLines( file.path( system.file( .NOZZLE.JAVASCRIPT.PATH, package=pkgname, lib.loc=libname ), .NOZZLE.TABLESORTER.MIN.FILE ), warn=FALSE ), env=.nozzleEnv );	
	
	assign( "googleAnalyticsCode", readLines( file.path( system.file( .NOZZLE.JAVASCRIPT.PATH, package=pkgname, lib.loc=libname ), .NOZZLE.GOOGLEANALYTICS.FILE ), warn=FALSE ), env=.nozzleEnv );	

	assign( ".PACKAGE", pkgname, env=.nozzleEnv );
}



# --- utility functions ----------------------------------------------------------------------------

.updateEnvCounter <- function( counter, increment=1, environment=.nozzleEnv )
{
	assign( counter, get( counter, env=.nozzleEnv ) + increment, env=environment );
	
	return ( get( counter, env=environment )  );
}

.initEnvCounter <- function( counter, environment=.nozzleEnv )
{
	assign( counter, 0, env=environment );
}

.getNextElementId <- function()
{
	return ( .updateEnvCounter( ".ELEMENT.ID" ) );
}


.setGlobalReportId <- function( id, environment=.nozzleEnv )
{
	assign( ".REPORT.ID", id, env=environment );	
}


.getGlobalReportId <- function( environment=.nozzleEnv )
{
	return ( get( ".REPORT.ID", environment ) );	
}


.getPackageVersion <- function( environment=.nozzleEnv )
{
	return ( .concat( get( ".PACKAGE", environment ), "_", .PACKAGE.VERSION ) );	

	# set lib.loc!
	#return ( .concat( get( ".PACKAGE", environment ), " ", packageVersion( get( ".PACKAGE", environment ), lib.loc=??? ) ) );	
}


.initNumbers <- function()
{
	.initEnvCounter( ".FIGURE.COUNTER" );
	.initEnvCounter( ".TABLE.COUNTER" );
	.initEnvCounter( ".FIGURE.SUPPLEMENT.COUNTER" );
	.initEnvCounter( ".TABLE.SUPPLEMENT.COUNTER" );
	.initEnvCounter( ".CITATION.COUNTER" );
	.initEnvCounter( ".RESULT.COUNTER" );
}


.getNextFigureNumber <- function()
{
	return ( .updateEnvCounter( ".FIGURE.COUNTER" ) );
}

.getNextTableNumber <- function()
{
	return ( .updateEnvCounter( ".TABLE.COUNTER" ) );
}

# supplementary figure numbers are negative
.getNextSupplementFigureNumber <- function()
{
	return ( .updateEnvCounter( ".FIGURE.SUPPLEMENT.COUNTER", increment=-1 ) );
}

# supplementary table numbers are negative
.getNextSupplementTableNumber <- function()
{
	return ( .updateEnvCounter( ".TABLE.SUPPLEMENT.COUNTER", increment=-1 ) );
}


.getNextCitationNumber <- function()
{
	return ( .updateEnvCounter( ".CITATION.COUNTER" ) );
}

.getNextResultId <- function()
{
	return ( .updateEnvCounter( ".RESULT.ID" ) );
}

.getNextResultNumber <- function()
{
	return ( .updateEnvCounter( ".RESULT.COUNTER" ) );
}


.newReportId <- function()
{
	# inspired by http://www.broofa.com/Tools/Math.uuid.js
	# creates an almost RFC 4122-compliant version 4 UUID (see Note 1 below)
	# Note 1: since this UUID is going to be part of a JavaScript function name dashes are replaced with underscores
	# Note 2: this is written for convenience, not performance
	# Note 3: the "R" at position 1 is required to make the id a valid HTML id
	
	template <- unlist( strsplit( "Rxxxxxxxx_xxxx_4xxx_yxxx_xxxxxxxxxxxx", "" ) );
	hex <- unlist( strsplit( "0123456789ABCDEF", "" ) );
	
	result <- unlist( lapply( template, function( c ) {	
				  r <- floor( runif( n=1, min=1, max=16 ) );
				  v <- 0;
				  
				  if ( c == "x" )
				  {
					v <- r;
				  }
				  else if ( c == "y" )
				  {
				  	v <- ( r %% 4 ) + 8;
				  }
				  else
				  {
					return ( c );
				  }
				  
				  return ( hex[v] );
				}) );

	return ( paste( result, sep="", collapse="" ) );
}


.removeProtectedElements <- function( element, level )
{
	if ( element$protection > level )
	{
		return ( NULL );
	}
	
	if ( length( element$elements ) > 0 )
	{	
		retainedElements <- list();
		
		for ( i in 1:length( element$elements ) )
		{
			result <- .removeProtectedElements( element$elements[[i]], level );
			
			if ( !is.null( result ) )
			{
				retainedElements[[length( retainedElements ) + 1]] <- result;
			}						
		}
		
		element$elements <- retainedElements;
	}
	
	return ( element );
}


.updateNumberedElements <- function( element, supplement=FALSE )
{
	if ( element$type == .ELEMENT.FIGURE )
	{
		if ( !supplement )
		{
			element$number <- .getNextFigureNumber();
		}
		else
		{
			element$number <- .getNextSupplementFigureNumber();		
		}
	}

	if ( element$type == .ELEMENT.TABLE )
	{
		if ( !supplement )
		{
			element$number <- .getNextTableNumber();
		}
		else
		{
			element$number <- .getNextSupplementTableNumber();		
		}
	}

	if ( element$type == .ELEMENT.CITATION )
	{
		element$number <- .getNextCitationNumber();
	}
	
	if ( length( element$elements ) > 0 )
	{	
		for ( i in 1:length( element$elements ) )
		{
			if ( !supplement && ( element$type == .ELEMENT.RESULT ) )
			{
				supplement <- TRUE;
			}
			
			element$elements[[i]] <- .updateNumberedElements( element$elements[[i]], supplement=supplement );
		}
	}
	
	return ( element );
}


.countSignificantResults <- function( element )
{
	counter <- 0;
	
	if ( element$type == .ELEMENT.RESULT )
	{
		if ( element$isSignificant == TRUE )
		{
		  counter <- counter + 1;
		}
	}
	
	if ( length( element$elements ) > 0 )
	{	
		for ( i in 1:length( element$elements ) )
		{
			counter <- counter + .countSignificantResults( element$elements[[i]] );
		}
	}
	
	return ( counter );
}


.getReference <- function( referenceId, element )
{
	if ( element$id == referenceId )
	{
		if ( element$type == .ELEMENT.CITATION )
		{
			return ( .concat( "[", element$number, "]" ) );		
		}
		
		return ( .concat( element$niceType, " ", element$number ) );
	}
		
	if ( length( element$elements ) > 0 )
	{	
		for ( i in 1:length( element$elements ) )
		{
			result <- .getReference( referenceId, element$elements[[i]] );
			
			if ( !is.null( result ) )
			{
				return ( result );
			}				
		}
	}
	
	return ( NULL );
}


.updateReferences <- function( element, report )
{
	if ( !is.null( element$text ) )
	{
		if ( is.list( element$text ) )
		{
			updatedText = "";
			
			if ( length( element$text ) > 0 )
			{
				for ( i in 1:length( element$text ) )
				{
					if ( is.list( element$text[[i]] ) || is.null( element$text[[i]] ) || is.na( element$text[[i]] ) )
					{
						next();
					}
					
					if ( substr( element$text[[i]], 0, nchar( .REFERENCE.STRING ) ) == .REFERENCE.STRING )
					{
						referenceId <- substr( element$text[[i]], nchar( .REFERENCE.STRING ) + 1, nchar( element$text[[i]] ) );
						referenceText <- .getReference( referenceId, report );
						
						if ( is.null( referenceText ) )
						{
							stop( paste( "Element", referenceId, "is referenced but not included in the report." ) );
						}
						
						updatedText <- paste( updatedText, referenceText, sep="" );
					}
					else
					{
						updatedText <-paste( updatedText, element$text[[i]], sep="" );
					}								
				}			
			}
			
			element$text <- updatedText;
		}
	}
	
	if ( length( element$elements ) > 0 )
	{	
		for ( i in 1:length( element$elements ) )
		{
			element$elements[[i]] <- .updateReferences( element$elements[[i]], report );
		}
	}		

	return ( element );
}


.updateResults <- function( element, report )
{
	# keeps track of all results that are encountered in this element
	resultList <- list();

	if ( !is.null( element$text ) )
	{
		if ( is.list( element$text ) )
		{
			updatedList <- list();
			j <- 1;
			
			if ( length( element$text ) > 0 )
			{
				for ( i in 1:length( element$text ) )
				{
					if ( is.list( element$text[[i]] ) )
					{
						# TODO: check if this is really of type result!!!!
						result <- element$text[[i]];
						
						if ( result$isSignificant == TRUE )
						{					
							if ( length( result$elements ) == 0 )
							{								
								#updatedList[[j]] <- .tag( "span", id=.concat( "resultid_", result$resultId ), class=c( "result", "significant", "noevidence" ) );
								updatedList[[j]] <- .tag( "span", class=c( "result", "significant", "noevidence" ) );
							}
							else
							{
								updatedList[[j]] <- .tag( "span", id=.concat( "resultid_", result$resultId ), class=c( "result", "significant" ) );
							}							
						}
						else
						{
							if ( length( result$elements ) == 0 )
							{
								#updatedList[[j]] <- .tag( "span", id=.concat( "resultid_", result$resultId ), class=c( "result", "noevidence" ) );
								updatedList[[j]] <- .tag( "span", class=c( "result", "noevidence" ) );
							}
							else
							{
								updatedList[[j]] <- .tag( "span", id=.concat( "resultid_", result$resultId ), class=c( "result" ) );
							}
						}
						j <- j + 1;
						
						result <- .updateReferences( result, report );
						
						updatedList[[j]] <- result$title;
						j <- j + 1;
						
						updatedList[[j]] <- .tag( "/span" );
						j <- j + 1;	
						
						# TODO: check if this is correct. the result should be added to the parent
						# add the result linked from the result summary to this element:
						#element$elements[[length( element$elements ) + 1]] <- result;
						resultList[[length( resultList ) + 1]] <- result;
					}
					else
					{
						updatedList[[j]] <-element$text[[i]];
						j <- j + 1;
					}				
				}			
			}
			
			element$text <- updatedList;
		}
	}
	
	if ( length( element$elements ) > 0 )
	{	
		updatedElements <- list();
		
		for ( i in 1:length( element$elements ) )
		{
			update <- .updateResults( element$elements[[i]], report );
			
			updatedElements[[length( updatedElements ) + 1]] <- update$element;
			
			# if results were found in the element they will be inserted after the element into the report tree
			if ( length( update$results ) > 0 )
			{
				for ( r in 1:length( update$results ) )
				{
					updatedElements[[length( updatedElements ) + 1]] <- update$results[[r]];
				}
			}
		}
		
		# replace old elements with updated list
		element$elements <- updatedElements;
	}		

	return ( list( element=element, results=resultList ) );
}


.log <- function( ..., severity=.SEVERITY.LOG )
{			
	message <- .concat( ... );
	
	if ( severity == .SEVERITY.LOG )
	{
		message( "Log: ", message, "\n" );
	}
	
	if ( severity == .SEVERITY.WARNING )
	{
		warning( message );
	}
	
	if ( severity == .SEVERITY.ERROR )
	{
		stop( message );
	}
}


# --- element construction -------------------------------------------------------------------------

.newElement <- function( elementType, title=NULL, text=NULL, niceType=NULL, domId=NULL, class=NULL, exportId=NULL, protection=PROTECTION.PUBLIC )
{
	element <- list( type = elementType );
	element$elements <- list();
	element$protection <- protection;
	
	element$id <- .getNextElementId();

	if ( !missing( title ) )
	{
		element$title <- title;
	}
	
	if ( !missing( text ) )
	{	
		element$text <- text;
	}
	
	if ( !missing( niceType ) )
	{	
		element$niceType <- niceType;
	}
	
	if ( !missing( domId ) )
	{
		element$domId <- domId;
	}

	if ( !missing( class ) )
	{
		element$class <- class;
	}

	if ( !missing( exportId ) )
	{
		element$exportId <- exportId;
	}
	else
	{
		element$exportId = NULL;
	}

	
	return ( element );
}


newReport <- function( ..., version=0 )
{
	if ( !missing( ... ) )
	{
		element <- .newGeneralReport( ..., version=version );
	}
	else
	{
		element <- .newGeneralReport( "Nozzle Report", version=version );
	}
	
	overview <- newSection( "Overview" );
	overview$domId <- "overview";

	introduction <- newSubSection( "Introduction" );
	introduction$domId <- "introduction";

	summary <- newSubSection( "Summary" );
	summary$domId  <- "summary";

	results <- newSection( "Results", class="results" );
	results$domId  <- "results";

	methods <- newSection( "Methods & Data" );
	methods$domId <- "methods";	

	input <- newSubSection( "Input" );
	input$domId <- "input";

	references <- newSubSection( "References" )
	references$domId <- "references";
	
	meta <- newSection( "Meta Information", class="meta" )
	meta$domId <- "meta";	
	
	# If a new predefined section is added here a ".hasPredefinedXXXSection" function needs to be added below as well as an an "addToXXX" function.
	
	element$predefined <- list();
	element$predefined$overviewId <- overview$id;
	element$predefined$introductionId <- introduction$id;
	element$predefined$summaryId <- summary$id;
	element$predefined$resultsId <- results$id;
	element$predefined$methodsId <- methods$id;
	element$predefined$inputId <- input$id;
	element$predefined$referencesId <- references$id;
	element$predefined$metaId <- meta$id;
	
	element <- addTo( parent=element, addTo( parent=overview, introduction, summary ) );
	element <- addTo( parent=element, results );
	element <- addTo( parent=element, addTo( parent=methods, input, references ) );
	element <- addTo( parent=element, meta );
}


.hasPredefinedOverviewSection <- function( report )
{
	return ( ( is.list( report$predefined ) && ( "overviewId" %in% names( report$predefined ) ) ) );
}


.hasPredefinedSummarySection <- function( report )
{
	return ( ( is.list( report$predefined ) && ( "summaryId" %in% names( report$predefined ) ) ) );
}


.hasPredefinedIntroductionSection <- function( report )
{
	return ( ( is.list( report$predefined ) && ( "introductionId" %in% names( report$predefined ) ) ) );
}


.hasPredefinedResultsSection <- function( report )
{
	return ( ( is.list( report$predefined ) && ( "resultsId" %in% names( report$predefined ) ) ) );
}


.hasPredefinedInputSection <- function( report )
{
	return ( ( is.list( report$predefined ) && ( "inputId" %in% names( report$predefined ) ) ) );
}


.hasPredefinedReferencesSection <- function( report )
{
	return ( ( is.list( report$predefined ) && ( "referencesId" %in% names( report$predefined ) ) ) );
}


.hasPredefinedMethodsSection <- function( report )
{
	return ( ( is.list( report$predefined ) && ( "methodsId" %in% names( report$predefined ) ) ) );
}


.hasPredefinedMetaSection <- function( report )
{
	return ( ( is.list( report$predefined ) && ( "metaId" %in% names( report$predefined ) ) ) );
}


newCustomReport <- function( ..., version=0 )
{
	if ( !missing( ... ) )
	{
		element <- .newGeneralReport( ..., version=version );
	}
	else
	{
		element <- .newGeneralReport( "Nozzle Custom Report", version=version );
	}	
}


.newGeneralReport <- function( ..., version=0 )
{
	element <- .newElement( .ELEMENT.REPORT, .concat( ... ) );
	
	element$meta <- list();
	
	# report version
	element$meta$version <- version;
	
	# report creator software
	element$meta$creator$name <- .getPackageVersion();
	element$meta$renderer$name <- element$meta$creator$name;

	# report creation date
	element$meta$creator$date <- date();		
	element$meta$renderer$date <- element$meta$creator$date;
	
	# maintainer details
	element$meta$maintainer$name <- NA;
	element$meta$maintainer$email <- NA;
	element$meta$maintainer$affiliation <- NA;	
		
	# report navigation: parent, prev, next
	element$navigation <- list();
	
	element$navigation$parentUrl <- NA;
	element$navigation$parentName <- NA;

	element$navigation$nextUrl <- NA;
	element$navigation$nextName <- NA;

	element$navigation$previousUrl <- NA;
	element$navigation$previousName <- NA;


	return ( element );
}


setNextReport <- function( report, url, ... )
{
	if ( is.null( report$navigation ) )
	{
		report$navigation <- list();
	}

	report$navigation$nextUrl <- url;
	report$navigation$nextName <- .concat( ... );	
	
	return ( report );
}


setPreviousReport <- function( report, url, ... )
{
	if ( is.null( report$navigation ) )
	{
		report$navigation <- list();
	}

	report$navigation$previousUrl <- url;
	report$navigation$previousName <- .concat( ... );	
	
	return ( report );
}

setParentReport <- function( report, url, ... )
{
	if ( is.null( report$navigation ) )
	{
		report$navigation <- list();
	}

	report$navigation$parentUrl <- url;
	report$navigation$parentName <- .concat( ... );	
	
	return ( report );
}


getReportTitle <- function( report )
{
	if ( is.null( report$title ) )
	{
		return ( NA );
	}

	return ( report$title );
}

setReportTitle <- function( report, title )
{
	if ( !is.null( report$title ) )
	{
		report$title <- title;
		
		return ( report );
	}

	return ( NA );
}


getReportId <- function( report )
{
	if ( is.null( report$reportId ) )
	{
		return ( NA );
	}

	return ( report$reportId );
}


# name and version of the Nozzle package that rendered the report
getRendererName <- function( report )
{
	if ( is.null( report$meta$renderer ) || is.atomic( report$meta$renderer ) )
	{
		return ( NA );
	}

	if ( is.null( report$meta$renderer$name ) )
	{
		return ( NA );
	}

	return ( report$meta$renderer$name );
}

# date on which the report was rendered/written to file
getRendererDate <- function( report )
{
	if ( is.null( report$meta$renderer ) || is.atomic( report$meta$renderer ) )
	{
		return ( NA );
	}

	if ( is.null( report$meta$renderer$date ) )
	{
		return ( NA );
	}

	return ( report$meta$renderer$date );
}


# name and version of the Nozzle package that was used to create the report
getCreatorName <- function( report )
{
	if ( is.null( report$meta$creator ) || is.atomic( report$meta$creator ) )
	{
		return ( NA );
	}

	if ( is.null( report$meta$creator$name ) )
	{
		return ( NA );
	}

	return ( report$meta$creator$name );
}


# date on which the report element was created
getCreatorDate <- function( report )
{
	if ( is.null( report$meta$creator ) || is.atomic( report$meta$creator ) )
	{
		return ( NA );
	}

	if ( is.null( report$meta$creator$date ) )
	{
		return ( NA );
	}

	return ( report$meta$creator$date );
}



# name of the software that generated the report
setSoftwareName <- function( report, ... )
{
	if ( is.null( report$meta$software ) )
	{
		report$meta$software <- list();
	}
	
	report$meta$software$name <- .concat( ... );	

	return ( report );
}


getSoftwareName <- function( report )
{
	if ( is.null( report$meta$software ) )
	{
		return ( NA );
	}

	if ( is.null( report$meta$software$name ) )
	{
		return ( NA );
	}

	return ( report$meta$software$name );
}


# version of the software that generated the report
setSoftwareVersion <- function( report, ... )
{
	if ( is.null( report$meta$software ) )
	{
		report$meta$software <- list();
	}
	
	report$meta$software$version <- .concat( ... );	

	return ( report );
}


getSoftwareVersion <- function( report )
{
	if ( is.null( report$meta$software ) )
	{
		return ( NA );
	}

	if ( is.null( report$meta$software$version ) )
	{
		return ( NA );
	}

	return ( report$meta$software$version );
}


# set logos for top{left|center|right} and bottom{left|center|right} of the report
setLogo <- function( report, filename, position )
{
	if ( is.null( report$meta$logo ) )
	{
		report$meta$logo <- list();
	}
	
	# top logos
	if ( position == LOGO.TOP.LEFT )
	{	
		report$meta$logo$topleft <- filename;	
	}
	
	if ( position == LOGO.TOP.CENTER )
	{	
		report$meta$logo$topcenter <- filename;	
	}
	
	if ( position == LOGO.TOP.RIGHT )
	{	
		report$meta$logo$topright <- filename;	
	}

	# bottom logos
	if ( position == LOGO.BOTTOM.LEFT )
	{	
		report$meta$logo$bottomleft <- filename;	
	}
	
	if ( position == LOGO.BOTTOM.CENTER )
	{	
		report$meta$logo$bottomcenter <- filename;	
	}
	
	if ( position == LOGO.BOTTOM.RIGHT )
	{	
		report$meta$logo$bottomright <- filename;	
	}

	return ( report );
}


# get logos for top{left|center|right} and bottom{left|center|right} of the report
getLogo <- function( report, position )
{
	if ( is.null( report$meta$logo ) )
	{
		return ( NA )
	}
	
	# top logos
	if ( position == LOGO.TOP.LEFT )
	{	
		if ( is.null( report$meta$logo$topleft ) )
		{
			return ( NA );
		}
		
		return ( report$meta$logo$topleft )
	}
	
	if ( position == LOGO.TOP.CENTER )
	{	
		if ( is.null( report$meta$logo$topcenter ) )
		{
			return ( NA );
		}
		
		return ( report$meta$logo$topcenter )
	}
	
	if ( position == LOGO.TOP.RIGHT )
	{	
		if ( is.null( report$meta$logo$topright ) )
		{
			return ( NA );
		}
		
		return ( report$meta$logo$topright )
	}

	# bottom logos
	if ( position == LOGO.BOTTOM.LEFT )
	{	
		if ( is.null( report$meta$logo$bottomleft ) )
		{
			return ( NA );
		}
		
		return ( report$meta$logo$bottomleft )
	}
	
	if ( position == LOGO.BOTTOM.CENTER )
	{	
		if ( is.null( report$meta$logo$bottomcenter ) )
		{
			return ( NA );
		}
		
		return ( report$meta$logo$bottomcenter )
	}
	
	if ( position == LOGO.BOTTOM.RIGHT )
	{	
		if ( is.null( report$meta$logo$bottomright ) )
		{
			return ( NA );
		}
		
		return ( report$meta$logo$bottomright )
	}


	return ( report );
}


# name of the person maintaining the report
setMaintainerName <- function( report, ... )
{
	if ( is.null( report$meta$maintainer ) )
	{
		report$meta$maintainer <- list();
	}
	
	report$meta$maintainer$name <- .concat( ... );	

	return ( report );
}

getMaintainerName <- function( report )
{
	if ( is.null( report$meta$maintainer$name ) )
	{
		return ( NA );
	}

	return ( report$meta$maintainer$name );
}

setMaintainerEmail <- function( report, ... )
{
	if ( is.null( report$meta$maintainer ) )
	{
		report$meta$maintainer <- list();
	}
	
	report$meta$maintainer$email <- .concat( ... );	

	return ( report );
}


getMaintainerEmail <- function( report )
{
	if ( is.null( report$meta$maintainer$email ) )
	{
		return ( NA );
	}

	return ( report$meta$maintainer$email );
}


setMaintainerAffiliation <- function( report, ... )
{
	if ( is.null( report$meta$maintainer ) )
	{
		report$meta$maintainer <- list();
	}
	
	report$meta$maintainer$affiliation <- .concat( ... );	

	return ( report );
}


getMaintainerAffiliation <- function( report )
{
	if ( is.null( report$meta$maintainer$affiliation ) )
	{
		return ( NA );
	}

	return ( report$meta$maintainer$affiliation );
}

setCopyright <- function( report, owner, year, statement=NA, url=NA )
{
	if ( is.null( report$meta$copyright ) )
	{
		report$meta$copyright <- list();
	}
	
	report$meta$copyright$owner <- owner;	
	report$meta$copyright$year <- year;		
	report$meta$copyright$statement <- statement;
	report$meta$copyright$url <- url;		

	return ( report );
}


setContactInformation <- function( report, email, subject, message, label=NA )
{
	if ( is.null( report$meta$contact ) )
	{
		report$meta$contact <- list();
	}
	
	report$meta$contact$label <- label;	
	report$meta$contact$email <- email;	
	report$meta$contact$subject <- subject;	
	report$meta$contact$message <- gsub( "\n", "%0A", message ); # line breaks will break the Javascript code

	return ( report );
}

getContactInformationEmail <- function( report )
{
	if ( is.null( report$meta$contact$email ) )
	{
		return ( NA );
	}

	return ( report$meta$contact$email );
}

getContactInformationSubject <- function( report )
{
	if ( is.null( report$meta$contact$subject ) )
	{
		return ( NA );
	}

	return ( report$meta$contact$subject );
}

getContactInformationMessage <- function( report )
{
	if ( is.null( report$meta$contact$message ) )
	{
		return ( NA );
	}

	return ( report$meta$contact$message );
}

getContactInformationLabel <- function( report )
{
	if ( is.null( report$meta$contact$label ) )
	{
		return ( NA );
	}

	return ( report$meta$contact$label );
}

getCopyrightOwner <- function( report )
{
	if ( is.null( report$meta$copyright$owner ) )
	{
		return ( NA );
	}

	return ( report$meta$copyright$owner );
}

getCopyrightYear <- function( report )
{
	if ( is.null( report$meta$copyright$year ) )
	{
		return ( NA );
	}

	return ( report$meta$copyright$year );
}

getCopyrightUrl <- function( report )
{
	if ( is.null( report$meta$copyright$url) )
	{
		return ( NA );
	}

	return ( report$meta$copyright$url );
}


getCopyrightStatement <- function( report )
{
	if ( is.null( report$meta$copyright$statement ) )
	{
		return ( NA );
	}

	return ( report$meta$copyright$statement );
}


setSignificantEntity <- function( report, ... )
{
	report$meta$significantEntity <- .concat( ... );

	return ( report );
}


getSignificantEntity <- function( report )
{
	if ( is.null( report$meta$significantEntity ) )
	{
		return ( DEFAULT.SIGNIFICANT.ENTITIY );
	}

	return ( report$meta$significantEntity );
}


# Create a Google Analytics account. The "web property id" is what you should enter here (usually starts with "UA-").
setGoogleAnalyticsId <- function( report, id )
{
	report$meta$googleAnalyticsId <- id;	

	return ( report );
}


getGoogleAnalyticsId <- function( report )
{
	if ( is.null( report$meta$googleAnalyticsId ) )
	{
		return ( NA );
	}

	return ( report$meta$googleAnalyticsId );
}


# overwrite built-in CSS with custom screen CSS file - path can be relative or absolute
setCustomScreenCss <- function( report, cssFile )
{
	report$customScreenCss <- cssFile	

	return ( report );
}


getCustomScreenCss <- function( report )
{
	if ( is.null( report$customScreenCss ) )
	{
		return ( NA );
	}

	return ( report$customScreenCss );
}


# overwrite built-in CSS with custom print CSS file - path can be relative or absolute
setCustomPrintCss <- function( report, cssFile )
{
	report$customPrintCss <- cssFile	

	return ( report );
}


getCustomPrintCss <- function( report )
{
	if ( is.null( report$customPrintCss ) )
	{
		return ( NA );
	}

	return ( report$customPrintCss );
}


.addToElementById <- function( element, id, ... )
{	
	# this function will always return the element if successful or not
	# test first with .getElementById (see below) if the element exists
	
	if ( element$id == id )
	{
		element <- addTo( parent=element, ... );
	}
		
	if ( length( element$elements ) > 0 )
	{	
		for ( i in 1:length( element$elements ) )
		{
			element$elements[[i]] <- .addToElementById( element$elements[[i]], id, ... );
		}
	}
	
	return ( element );	
}

.getElementById <- function( element, id )
{	
	if ( element$id == id )
	{
		return ( element );
	}
		
	if ( length( element$elements ) > 0 )
	{	
		for ( i in 1:length( element$elements ) )
		{
			result <- .getElementById( element$elements[[i]], id );
			
			if ( !is.null( result ) )
			{
				return ( result );
			}
		}
	}
	
	return ( NULL );	
}

# removes element the given id from the tree using breadth-first search
.removeElementById <- function( element, id )
{	
	if ( length( element$elements ) > 0 )
	{	
		newElements <- list();
		
		for ( i in 1:length( element$elements ) )
		{
			if ( element$elements[[i]]$id != id )
			{
				newElements[[length( newElements ) + 1]] <- element$elements[[i]];								
			}
		}
		
		element$elements <- newElements;
		
		if ( length( element$elements ) > 0 )
		{			
			for ( i in 1:length( element$elements ) )
			{
				element$elements[[i]] <- .removeElementById( element$elements[[i]], id );
			}		
		}
	}
	
	return ( element );	
}



getExportedElement <- function( report, exportId )
{	
	if ( !is.null( report$exportId ) && ( report$exportId == exportId ) )
	{
		return ( report );
	}
		
	if ( length( report$elements ) > 0 )
	{	
		for ( i in 1:length( report$elements ) )
		{
			result <- getExportedElement( report$elements[[i]], exportId );
			
			if ( !is.null( result ) )
			{
				return ( result );
			}
		}
	}
	
	return ( NULL );	
}


getExportedElementIds <- function( report )
{	
	if ( !is.null( report$exportId ) )
	{
		return( report$exportId );
	}
	
	elementIds <- NULL;	
					
	if ( length( report$elements ) > 0 )
	{	
		for ( i in 1:length( report$elements ) )
		{
			elementIds <- c( elementIds, getExportedElementIds( report$elements[[i]] ) );			
		}
	}
	
	return ( unique( elementIds ) );	
}


addToResults <- function( report, ... )
{
	if ( .hasPredefinedResultsSection( report ) )
	{
		return ( .addToElementById( report, report$predefined$resultsId, ... ) ); 
	}
	
	stop( "This report does not contain a predefined results section. Is this a custom report?" );
}


addToMethods <- function( report, ... )
{
	if ( .hasPredefinedMethodsSection( report ) )
	{
		return ( .addToElementById( report, report$predefined$methodsId, ... ) ); 
	}
	
	stop( "This report does not contain a predefined methods section. Is this a custom report?" );
}


addToOverview <- function( report, ... )
{
	if ( .hasPredefinedOverviewSection( report ) )
	{
		return ( .addToElementById( report, report$predefined$overviewId, ... ) ); 
	}

	stop( "This report does not contain a predefined overview section. Is this a custom report?" );
}


addToIntroduction <- function( report, ... )
{
	if ( .hasPredefinedIntroductionSection( report ) )
	{
		return ( .addToElementById( report, report$predefined$introductionId, ... ) ); 
	}

	stop( "This report does not contain a predefined introduction section. Is this a custom report?" );
}

addToSummary <- function( report, ... )
{
	if ( .hasPredefinedSummarySection( report ) )
	{
		return ( .addToElementById( report, report$predefined$summaryId, ... ) ); 
	}

	stop( "This report does not contain a predefined summary section. Is this a custom report?" );
}

addToInput <- function( report, ... )
{
	if ( .hasPredefinedInputSection( report ) )
	{
		return ( .addToElementById( report, report$predefined$inputId, ... ) ); 
	}

	stop( "This report does not contain a predefined input section. Is this a custom report?" );
}

addToReferences <- function( report, ... )
{
	if ( .hasPredefinedReferencesSection( report ) )
	{
		return ( .addToElementById( report, report$predefined$referencesId, ... ) ); 
	}

	stop( "This report does not contain a predefined references section. Is this a custom report?" );
}

addToMeta <- function( report, ... )
{
	if ( .hasPredefinedMetaSection( report ) )
	{
		return ( .addToElementById( report, report$predefined$metaId, ... ) ); 
	}

	stop( "This report does not contain a predefined meta information section. Is this a custom report?" );
}


newSection <- function( ..., class="", exportId=NULL, protection=PROTECTION.PUBLIC )
{
	return ( .newElement( .ELEMENT.SECTION, .concat( ... ), class=class, exportId=exportId, protection=protection ) );
}


newSubSection <- function( ..., exportId=NULL, protection=PROTECTION.PUBLIC )
{
	return ( .newElement( .ELEMENT.SUBSECTION, .concat( ... ), exportId=exportId, protection=protection ) );
}


newSubSubSection <- function( ..., exportId=NULL, protection=PROTECTION.PUBLIC )
{
	return ( .newElement( .ELEMENT.SUBSUBSECTION, .concat( ... ), exportId=exportId, protection=protection ) );
}


.writeSection <- function ( element, file, level="" )
{
	if ( is.null( element$domId ) )
	{	
		.write( .tag( "div", class=.concat( level, "section", " ", element$class ) ), file );
	}
	else
	{
		.write( .tag( "div", class=.concat( level, "section", " ", element$class ), id=element$domId ), file );
	}

	.write( .tag( "div", class="sectionheader" ), file );
	.write( element$title, file );
	.write( .tag( "/div" ), file );
	
	.write( .tag( "div", class="sectionbody" ), file );
	.writeElements( element, file );
	.write( .tag( "/div" ), file );

	.write( .tag( "/div" ), file );	
}


newList <- function( ..., isNumbered=FALSE, exportId=NULL, protection=PROTECTION.PUBLIC )
{
	element <- .newElement( .ELEMENT.LIST, exportId=exportId, protection=protection );
	
	element$isNumbered <- isNumbered;
	
	if ( missing( ... ) )
	{
		return ( element );
	}	
	
	args <- list( ... );
	
	# TODO: run the same checks in "addTo"
	for ( i in 1:length( args ) )
	{
		if ( args[[i]]$type == .ELEMENT.PARAGRAPH ||
			args[[i]]$type == .ELEMENT.LIST ||
			args[[i]]$type == .ELEMENT.RESULT )
		{
			element$elements[[length( element$elements ) + 1]] <- args[[i]];
		}
		else
		{
			stop( "Report elements of type ", args[[i]]$element$type, " can not be added to a list." );
		}
	}
	
	return ( element );
}

.writeList <- function( element, file )
{
	if ( element$isNumbered == TRUE )
	{
		.write( .tag( "ol" ), file );	
	}
	else
	{
		.write( .tag( "ul" ), file );			
	}
	
	for ( i in 1:length( element$elements ) )
	{
		if ( ( element$elements[[i]]$type != .ELEMENT.LIST ) && ( element$elements[[i]]$type != .ELEMENT.RESULT ) )
		{
			.write( .tag( "li" ), file );		
			.writeElement( element$elements[[i]], file );		
			.write( .tag( "/li" ), file );
		}
		else
		{		
			.writeElement( element$elements[[i]], file );		
		}
	}

	if ( element$isNumbered == TRUE )
	{
		.write( .tag( "/ol" ), file );	
	}	
	else
	{
		.write( .tag( "/ul" ), file );			
	}
}


newFigure <- function( file, ..., fileHighRes=NA, type=IMAGE.TYPE.RASTER, exportId=NULL, protection=PROTECTION.PUBLIC )
{
	element <- .newElement( .ELEMENT.FIGURE, niceType="Figure", exportId=exportId, protection=protection );
	
	element$imageFilename <- file;
	element$imageFilenameHighRes <- fileHighRes;
	element$imageType <- type;
	
	element$text <- list( ... );
	
	return ( element );
}

isFigure <- function( element )
{
	return ( element$type == .ELEMENT.FIGURE );
}

getFigureFile <- function( element )
{
	if ( isFigure( element ) )
	{
		return ( element$imageFilename );
	}
	
	return ( NA );
}

setFigureFile <- function( element, file )
{
	if ( isFigure( element ) )
	{
		element$imageFilename <- file;
		
		return ( element );
	}
	
	return ( NA );
}


getFigureFileHighRes <- function( element )
{
	if ( isFigure( element ) )
	{
		return ( element$imageFilenameHighRes );
	}
	
	return ( NA );
}


setFigureFileHighRes <- function( element, file )
{
	if ( isFigure( element ) )
	{
		element$imageFilenameHighRes <- file;
		
		return ( element );
	}
	
	return ( NA );
}


.writeFigure <- function ( element, file )
{
	figureNumber <- element$number;
	
	# supplementary figure?
	if ( figureNumber < 0 )
	{
		figureNumber <- .concat( "S", abs( figureNumber ) );
	}
	
	.write( .tag( "div", class="figure", id=.concat( "figure_", figureNumber ) ), file );

	.write(	.tag( "div", class="caption" ), file );
	.write( .tag( "p" ), file );
	
	# "Figure 1. " ...
	.write( asStrong( element$niceType, "&nbsp;", figureNumber, ".&nbsp;" ), file );
	
	if ( !is.na( element$imageFilenameHighRes ) )
	{
		.write( "\n", .tag( "a", class="download", other=.concat( "href=\"", element$imageFilenameHighRes , "\"" ) ), "Get High-res Image", .tag( "/a" ), "\n", file );	
	}
	.write( element$text, file );
	.write( .tag( "/p" ), file );
	.write( .tag( "/div" ), file ); # caption

	.write( .tag( "div", class="image" ), file );
	
	
	if ( ( element$imageType == IMAGE.TYPE.RASTER ) || ( is.null( element$imageType ) ) )
	{
		.write( .tag( "img", other=.concat( "src=\"", element$imageFilename, "\"/" ) ), file );
	}
	else if ( element$imageType == IMAGE.TYPE.SVG )
	{
		.write( .tag( "object", other=.concat( "data=\"", element$imageFilename, "\" type=\"image/svg+xml\"" ) ), file );
		.write( .tag( "/object" ), file );
	}
	else
	{
		error( "Unknown image type selected in newFigure()." );
	}		
	
	.write( .tag( "/div" ), file ); # image
																
	.write( .tag( "/div" ), file ); # figure
}


newTable <- function( table, ..., file=NA, significantDigits=TABLE.SIGNIFICANT.DIGITS, exportId=NULL, protection=PROTECTION.PUBLIC )
{
	element <- .newElement( .ELEMENT.TABLE, niceType="Table", exportId=exportId, protection=protection );
	
	element$table <- table;
	element$tableSignificantDigits <- significantDigits;
	element$tableFilename <- file;
	
	element$resultList <- list();
	element$resultIndices <- array( dim=dim( table ) );
	
	element$text <- list( ... );
	
	return ( element );
}


isTable <- function( element )
{
	return ( element$type == .ELEMENT.TABLE );
}

getTableFile <- function( element )
{
	if ( isTable( element ) )
	{
		return ( element$tableFilename );
	}
	
	return ( NA );
}

setTableFile <- function( element, file )
{
	if ( isTable( element ) )
	{
		element$tableFilename <- file;
		
		return ( element );
	}
	
	return ( NA );
}



.trimTable <- function( table, significantDigits=3 )
{
	if ( significantDigits <= 0 )
	{
		return ( table );
	}
	
	for( c in 1:dim( table )[2] )
	{		
		# if this column is numeric ...
		if ( is.numeric( table[,c] ) == TRUE )
		{			
			# if there are entries with decimal digits ...
			if ( any( trunc( table[,c] ) != table[,c], na.rm=TRUE ) )
			{
				# trim to significant digits
				table[,c] <- signif( table[,c], digits=significantDigits );
			}
			else
			{
				# add separators between every block of 3 digits
				# PROBLEM: TableSorter doesn't sort these numbers correctly
				#table[,c] <- format( table[,c], big.mark=",", scientific=FALSE );
			}
		}
	}
	
	return ( table );
}


.writeTable <- function ( element, file )
{	
	tableNumber <- element$number;
	
	# supplementary table?
	if ( tableNumber < 0 )
	{
		tableNumber <- .concat( "S", abs( tableNumber ) );
	}

	.write( .tag( "div", class="table", id=.concat( "table_", tableNumber ) ), file );
							
	.write(	.tag( "div", class="caption" ), file );
	.write( .tag( "p" ), file );
	
	# "Table 1. " ...
	.write( asStrong( element$niceType, "&nbsp;", tableNumber, ".&nbsp;" ), file );

	if ( !is.na( element$tableFilename ) )
	{
		.write( "\n", .tag( "a", class="download", other=.concat( "href=\"", element$tableFilename, "\"" ) ), "Get Full Table", .tag( "/a" ), "\n", file );	
	}

	.write( element$text, file );
	.write( .tag( "/p" ), file );
	.write( .tag( "/div" ), file ); # caption

	# write table only if a matrix/data frame has been supplied
	if ( is.null( dim( element$table ) ) )	
	{
		.write( .tag( "table", class="resulttable tablesorter sortabletable" ), file );
		
		# --- table header ---
	
		.write( .tag( "thead" ), file );
		.write( .tag( "tr" ), file );
		
		# write column names
		for( c in 1:length( element$table ) )
		{
			.write( .tag( "th" ), file );
			.write( element$table[c], file );
			.write( .tag( "/th" ), file );
		}
			
		.write( .tag( "/tr" ), file );
		.write( .tag( "/thead" ), file );		
		
		.write( .tag( "tbody" ), file );
		.write( .tag( "tr" ), file );
		.write( .tag( "td", other=.concat( "colspan=\"", length( element$table ), "\" align=\"center\"" ) ), file );
		.write( .tag( "a", other=.concat( "href=\"", element$tableFilename, "\"" ) ), "Get Full Table", .tag( "/a" ), file );		
		.write( .tag( "/td" ), file );
		.write( .tag( "/tr" ), file );
		.write( .tag( "/tbody" ), file );
		
		.write( .tag( "/table" ), file ); # resultstable			
	}
	else
	{
		# TODO: make sure that the resulttable class is attached to these tables
		#.write( .tag( "table", id=.concat( "sortable_table_", element$number ), class="tablesorter" ), file );
		#.write( .tag( "table", id=.concat( "sortable_table_", element$number ), class="resulttable tablesorter" ), file );
		.write( .tag( "table", class="resulttable tablesorter sortabletable" ), file );
		
		# --- table header ---
	
		.write( .tag( "thead" ), file );
		.write( .tag( "tr" ), file );
		
		# write column names
		for( c in 1:dim( element$table )[2] )
		{
			.write( .tag( "th" ), file );
			.write( colnames( element$table )[c], file );
			.write( .tag( "/th" ), file );
		}
			
		.write( .tag( "/tr" ), file );
		.write( .tag( "/thead" ), file );
		
		
		# --- table body ---
		
		# copy original table and perform trimming to significant digits	
		trimmedTable <- 0;
			
		if ( is.null( ( element$tableSignificantDigits ) ) )
		{
			trimmedTable <- .trimTable( element$table, TABLE.SIGNIFICANT.DIGITS );
		}
		else
		{	
			if ( is.na( element$tableSignificantDigits ) )
			{	
				trimmedTable <- .trimTable( element$table, TABLE.SIGNIFICANT.DIGITS );
			}
			else
			{
				trimmedTable <- .trimTable( element$table, element$tableSignificantDigits );
			}
		}
		
		.write( .tag( "tbody" ), file );
		
		for( r in 1:dim( element$table )[1] )
		{
			.write( .tag( "tr" ), file );
			
			for( c in 1:dim( element$table )[2] )
			{
				.write( .tag( "td" ), file );
				
				resultIndex <- element$resultIndices[r,c];
				
				if ( !is.na( resultIndex ) )
				{
					result <- element$resultList[[resultIndex]];
					
					if ( result$isSignificant )
					{					
						if ( length( result$elements ) == 0 )
						{
							#.write( .tag( "span", class=c( "result", "significant", "noevidence" ), id=.concat( "resultid_", result$resultId, "_", result$resultNumber ) ), file );
							.write( .tag( "span", class=c( "result", "significant", "noevidence" ) ), file );
						}
						else
						{	
							.write( .tag( "span", class=c( "result", "significant" ), id=.concat( "resultid_", result$resultId ) ), file );						
						}
					}
					else
					{
						if ( length( result$elements ) == 0 )
						{
							#.write( .tag( "span", class=c( "result", "noevidence" ), id=.concat( "resultid_", result$resultId, "_", result$resultNumber ) ), file );
							.write( .tag( "span", class=c( "result", "noevidence" ) ), file );
						}
						else
						{
							.write( .tag( "span", class=c( "result" ), id=.concat( "resultid_", result$resultId ) ), file );
						}
					}
				}
				
				#.write( as.character( element$table[r,c] ), file );
				#.write( as.character( trimmedTable[r,c] ), file );
				.write( format( trimmedTable[r,c], scientific=1 ), file );
				
				if ( !is.na( resultIndex ) )
				{
					.write( .tag( "/span" ), file );
				}
	
				.write( .tag( "/td" ), file );						
			}		
			
			.write( .tag( "/tr" ), file );
		}
			
		.write( .tag( "/tbody" ), file );
		
		.write( .tag( "/table" ), file ); # resultstable	
	}
	
	.write( .tag( "/div" ), file ); # table	

	# write all elements, i.e. results only in this case
	.writeElements( element, file );																		
}



newCitation <- function( authors="", title, publication="", issue="", number="", pages="", year="", url="" )
{
	element <- .newElement( .ELEMENT.CITATION, niceType="Citation" );
	
	element$authors <- authors;
	element$year <- year;
	element$title <- title;
	element$publication <- publication;
	element$publicationIssue <- issue;
	element$publicationNumber <- number;
	element$pages <- pages;
	element$url <- url;	
	
	return ( element );
}

newJournalCitation <- function( authors, title, publication, issue, number, pages, year, url="" )
{
	return ( newCitation( authors=authors, title=title, publication=publication, issue=issue, number=number, pages=pages, year=year, url=url ) );
}

newWebCitation <- function( authors, title, url )
{
	return ( newCitation( authors=authors, title=title, url=url ) );
}


.writeCitation <- function( element, file )
{
	.write( .tag( "div", class="reference", id=.concat( "reference_", element$number ) ), file );
	
	.write( "[", element$number, "] ", nobreak=TRUE, file );

	if ( element$authors != "" )
	{
		.write( element$authors, ", ", nobreak=TRUE, file );
	}

	if ( element$title != "" ) # only required field
	{
		.write( asLink( url=element$url, element$title ), nobreak=TRUE, file );
	}

	if ( element$publication != "" )
	{
		.write( ", ", asEmph( element$publication ), nobreak=TRUE, file );
	}

	if ( element$publication != "" && element$publicationIssue != "" )
	{
		.write( " ", asStrong( element$publicationIssue ), nobreak=TRUE, file );
	}

	if ( element$publication != "" && element$publicationIssue != "" && element$publicationNumber != "" )
	{
		.write( "(", element$publicationNumber, ")", nobreak=TRUE, file );
	}

	if ( element$pages != "" )
	{
		.write( ":", element$pages, nobreak=TRUE, file );
	}
	
	if ( element$year != "" )
	{
		.write( " (",element$year, ")", nobreak=TRUE, file );
	}
	
	.write( .tag( "/div" ), file ); # reference	
}


newParagraph <- function( ..., exportId=NULL, protection=PROTECTION.PUBLIC )
{
	element <- .newElement( .ELEMENT.PARAGRAPH, exportId=exportId, protection=protection );

	element$text <- list( ... );
	
	return ( element );
}


# does not need a write function, writing is provided by .writeList
newParameterList <- function( ..., separator=" = ", exportId=NULL, protection=PROTECTION.PUBLIC )
{
	element <- newList( isNumbered=FALSE, exportId=exportId, protection=protection );
	
	args <- list( ... );
	
	# if the only argument passed is a list then use it as the list of arguments
	if ( ( length( args ) == 1 ) && ( is.list( args[[1]] ) ) )
	{
		args <- args[[1]];
	}
	
	if ( length( args ) < 2 )
	{
		stop( "At least one parameter-value pair has to be provided to parameter list." );
	}
	
	if ( length( args ) %% 2 != 0 )
	{
		stop( "Number of elements in parameter list is not multiple of 2." );
	}
	
	j <- 1;
	
	for ( i in 1:(length( args )/2) )
	{
		element <- addTo( parent=element, newParagraph( args[[j]], separator, args[[j+1]] ) );
		
		j <- j + 2;
	}
	
	return ( element );
}


.writeParagraph <- function ( element, file )
{
	.write( .tag( "p" ), file );
	.write( element$text, file );
	
	# write all elements, i.e. results only in this case
	.writeElements( element, file );	
	
	.write( .tag( "/p" ), file );	
}


newResult <- function( ..., isSignificant=FALSE, protection=PROTECTION.PUBLIC )
{
	element <- .newElement( .ELEMENT.RESULT, .concat( ... ), protection=protection );

	element$resultId <- .getNextResultId();
	element$isSignificant <- isSignificant;
	
	return ( element );
}


.writeResult <- function ( element, file )
{
	# only write non-empty results
	if ( length( element$elements ) > 0 )
	{		
		.write( .tag( "div", class="evidence", id=.concat( "evidenceid_", element$resultId ) ), file );
	
		.writeElements( element, file );
	
		.write( .tag( "/div" ), file );		
	}	
}


.writeTopLogos <- function ( report, file )
{
	# quit if there are no logos to be added
	if ( is.na( getLogo( report, LOGO.TOP.LEFT ) ) &&
		 is.na( getLogo( report, LOGO.TOP.CENTER ) ) &&
		 is.na( getLogo( report, LOGO.TOP.RIGHT ) ) )
	{
		return;
	}

	# write top logos
	.write( .tag( "div", class="logo" ), file );
	.write( .tag( "div", class="top" ), file );
	
	.write( .tag( "div", class="left" ), file );
	if ( !is.na( getLogo( report, LOGO.TOP.LEFT ) ) )
	{	
		.write( .tag( "img", other=.concat( "src=\"", getLogo( report, LOGO.TOP.LEFT ), "\"/" ) ), file );
	}
	.write( .tag( "/div" ), file );

	.write( .tag( "div", class="center" ), file );
	if ( !is.na( getLogo( report, LOGO.TOP.CENTER ) ) )
	{	
		.write( .tag( "img", other=.concat( "src=\"", getLogo( report, LOGO.TOP.CENTER ), "\"/" ) ), file );
	}
	.write( .tag( "/div" ), file );

	.write( .tag( "div", class="right" ), file );
	if ( !is.na( getLogo( report, LOGO.TOP.RIGHT ) ) )
	{	
		.write( .tag( "img", other=.concat( "src=\"", getLogo( report, LOGO.TOP.RIGHT ), "\"/" ) ), file );
	}
	.write( .tag( "/div" ), file );
	
	.write( .tag( "/div" ), file );	
	.write( .tag( "/div" ), file );
}


.writeBottomLogos <- function ( report, file )
{
	# quit if there are no logos to be added
	if ( is.na( getLogo( report, LOGO.BOTTOM.LEFT ) ) &&
		 is.na( getLogo( report, LOGO.BOTTOM.CENTER ) ) &&
		 is.na( getLogo( report, LOGO.BOTTOM.RIGHT ) ) )
	{
		return;
	}


	# write bottom logos
	.write( .tag( "div", class="logo" ), file );
	.write( .tag( "div", class="bottom" ), file );
	
	.write( .tag( "div", class="left" ), file );
	if ( !is.na( getLogo( report, LOGO.BOTTOM.LEFT ) ) )
	{	
		.write( .tag( "img", other=.concat( "src=\"", getLogo( report, LOGO.BOTTOM.LEFT ), "\"/" ) ), file );
	}
	.write( .tag( "/div" ), file );

	.write( .tag( "div", class="center" ), file );
	if ( !is.na( getLogo( report, LOGO.BOTTOM.CENTER ) ) )
	{	
		.write( .tag( "img", other=.concat( "src=\"", getLogo( report, LOGO.BOTTOM.CENTER ), "\"/" ) ), file );
	}
	.write( .tag( "/div" ), file );

	.write( .tag( "div", class="right" ), file );
	if ( !is.na( getLogo( report, LOGO.BOTTOM.RIGHT ) ) )
	{	
		.write( .tag( "img", other=.concat( "src=\"", getLogo( report, LOGO.BOTTOM.RIGHT ), "\"/" ) ), file );
	}
	.write( .tag( "/div" ), file );
	
	.write( .tag( "/div" ), file );	
	.write( .tag( "/div" ), file );
}



addTo <- function( parent, ..., row=NA, column=NA )
{
	args <- list( ... );
	
	if ( parent$type == .ELEMENT.TABLE )
	{
		if ( !missing( row ) && !missing( column ) )
		{			
			if ( length( args ) == 1 )
			{
				if ( args[[1]]$type == .ELEMENT.RESULT )
				{
					parent <- .addResult( table=parent, row=row, column=column, result=args[[1]] );
					
					return ( parent );
				}
			}
		}
	}
	
	
	if ( length( args ) > 0 )
	{
		for ( i in 1:length( args ) )
		{
			parent$elements[[length( parent$elements ) + 1]] <- args[[i]];	
		}	
	}
		
	return ( parent );
}


.addResult <- function( table, row, column, result )
{
	# !!! do not change the order of the following two lines!!!
	table$resultList[[length( table$resultList ) + 1]] <- result;	
	table$resultIndices[row,column] = length( table$resultList );
	
	# add result right after table
	table$elements[[length( table$elements ) + 1]] <- result;	
	
	return ( table );
}



asReference <- function( element )
{
	# TODO: create link if requested
	# TODO: check for type!
	return ( paste( .REFERENCE.STRING, element$id, sep="" ) );
}

asSummary <- function( result )
{
	return ( result );
}


asLink <- function( url, ... )
{
	# TODO: open link in new window if requested		
	
	args <- list( ... );
	
	if ( missing( url ) && length( args ) == 0 )
	{
		stop( "Unable to create link without URL or link text." );
	}	
	
	if ( url == "" || missing( url ) )
	{
		return ( .concat( args ) );	# just print text but do not create a link
	}
	
	if ( length( args ) == 0 )
	{
		args <- url; # print URL as link text
	}
	
	# TODO: remove "\n"s
	#return ( .concat( "\n", .tag( "a", other=.concat( "href=\"", url, "\"" ) ), args, .tag( "/a" ), "\n" ) );
	return ( .concat( .tag( "a", other=.concat( "href=\"", url, "\"" ) ), args, .tag( "/a" ) ) );
}


#asIdentifier <- function( label, class=NA, column=NA, row=NA )
#{
#	if ( missing( class ) && missing( label ) )
#	{
#		stop( "Unable to create identifier without class or label." );
#	}	
#	
#	if ( missing( class ) )
#	{
#		return ( label ); # just print text
#	}
#	
#	return ( .concat( "<span class=\"identifier ", class, "\">", label, "</span>" ) );
#}


asStrong <- function( ... )
{
	return ( .concat( .tag( "strong" ), ..., .tag( "/strong" ) ) );
}


asEmph <- function( ... )
{
	return ( .concat( .tag( "em" ), ..., .tag( "/em" ) ) );
}


asParameter <- function( ... )
{
	return ( .concat( .tag( "span", class="parameter" ), ..., .tag( "/span" ) ) );
}


asValue <- function( ... )
{
	return ( .concat( .tag( "span", class="value" ), ..., .tag( "/span" ) ) );
}


asFilename <- function( ... )
{
	return ( .concat( .tag( "span", class="filename" ), ..., .tag( "/span" ) ) );
}


asCode <- function( ... )
{
	return ( .concat( .tag( "span", class="code" ), ..., .tag( "/span" ) ) );
}


.concat <- function( ... )
{
	if ( length( list( ... ) ) == 0 )
	{
		return ( "" );
	}
	
	return ( paste( ..., sep="" ) );
}


# helper function: writes a line of html code
.write <- function( ..., nobreak=FALSE, file )
{
  if ( missing( file ) )
  {
	args <- list( ... );
  	file <- args[[length( args )]];
  	args[[length( args )]] <- "";
  }  
  
  # string to be written to file
  string <- paste( args, sep="" );

  # if the report is supposed to be used as a fragment, all ids have to be made unique
  # scan text for this pattern "< ...id="xyz" ...>"  
  # replace with "< ...id="xyz_REPORTID" ...>"
  # sed s/"\(<.*\)id=\"\([A-Za-z][A-Za-z0-9_\.-\:]*\)\"\(.*>\)"/\\1id=\\2_TEXT\\3/g demo.html      
  if ( .getGlobalReportId() != "" )
  {
  	 string <- gsub( "(<.*)id=\"([A-Za-z][A-Za-z0-9_\\.-\\:]*)\"(.*>)", .concat( "\\1id=\"\\2_", .getGlobalReportId(), "\"\\3" ), string );  
  }	 
  
  if ( nobreak )
  {  
  	cat( string, file=file, sep="" );
  }
  else
  {
  	cat( string, "\n", file=file, sep="" );
  }
}


.tag <- function( tag, id=NA, class=NA, other=NA )
{
	idString <- "";
	classString <- "";
	otherString <- "";
	
	if ( !missing( id ) )
	{
		idString <- .concat( " id=\"", id, "\"" );	
	}

	if ( !missing( class ) )
	{
		classString <- .concat( " class=\"", paste( class, sep=" ", collapse=" " ), "\"" );	
	}
	
	if ( !missing( other ) )
	{
		otherString <- .concat( " ", other );
	}
	
	return ( .concat( "<", tag, classString, idString, otherString, ">" ) );
}

.comment <- function( ... )
{
	return ( .concat( "<!-- ", ..., " -->" ) );
}



.writeElement <- function( element, file )
{
	if ( element$type == .ELEMENT.PARAGRAPH )
	{
		.writeParagraph( element, file );
	}

	if ( element$type == .ELEMENT.SECTION )
	{
		.writeSection( element, file, "" );
	}

	if ( element$type == .ELEMENT.SUBSECTION )
	{
		.writeSection( element, file, "sub" );
	}

	if ( element$type == .ELEMENT.SUBSUBSECTION )
	{
		.writeSection( element, file, "subsub" );
	}

	if ( element$type == .ELEMENT.LIST )
	{
		.writeList( element, file );
	}

	if ( element$type == .ELEMENT.FIGURE )
	{
		.writeFigure( element, file );
	}

	if ( element$type == .ELEMENT.TABLE )
	{
		.writeTable( element, file );
	}

	if ( element$type == .ELEMENT.CITATION )
	{
		.writeCitation( element, file );
	}

	if ( element$type == .ELEMENT.RESULT )
	{
		.writeResult( element, file );
	}
}


.writeElements <- function( element, file )
{
	if ( length( element$elements ) > 0 )
	{	
		for ( i in 1:length( element$elements ) )
		{
			.writeElement( element$elements[[i]], file );
		}
	}
}


getSummary <- function( report )
{
	.initNumbers();

	return ( .updateReferences( .updateNumberedElements( .getElementById( report, report$predefined$summaryId ) ), report=report )$elements ); 	
}


getSignificantResultsCount <- function( report )
{
	return ( .countSignificantResults( report ) );
}


writeReport <- function( report, filename=DEFAULT.REPORT.FILENAME, debug=FALSE, output=c( HTML.REPORT, RDATA.REPORT ), credits=TRUE, level=PROTECTION.PUBLIC, debugCss=NA, debugJavaScript=NA )
{	
	# check if input and references contain any reports, if not, remove those sections		
	
	if ( is.list( report$predefined ) && ( "referencesId" %in% names( report$predefined ) ) )
	{	
		references <- .getElementById( report, report$predefined$referencesId );	
		
		if ( !is.null( references ) )
		{		
			# remove reference section
			report <- .removeElementById( report, references$id );
	
			if ( length( references$elements ) > 0 )
			{
				# there are some references, insert them at the end of the methods section
				report <- addToMethods( report, references );
			}					
		}
	}

	if ( is.list( report$predefined ) && ( "inputId" %in% names( report$predefined ) ) )	{	
		input <- .getElementById( report, report$predefined$inputId );
		
		if ( !is.null( input ) )
		{
			if ( length( input$elements ) == 0 )
			{
				# no input, remove the empty section
				report <- .removeElementById( report, input$id );
			}
		}
	}

	if ( is.list( report$predefined ) && ( "metaId" %in% names( report$predefined ) ) )
	{	
		meta <- .getElementById( report, report$predefined$metaId );
		
		if ( !is.null( meta ) )
		{
			if ( length( meta$elements ) == 0 )
			{
				# no meta information, remove the empty section
				report <- .removeElementById( report, meta$id );
			}
		}
	}	
	
	# --- additional meta information --- #
	
	# assign report id
	report$reportId <- .newReportId();
	
	# assign date of report writing (if doesn't exist yet)
	if ( is.null( report$meta$createDate ) )
	{
		report$meta$createDate <- date();
	}	

	# write as requested output types	
	if ( RDATA.REPORT %in% output )
	{		
		save( report, file=.concat( filename, ".RData" ) );		
	}
		
	if ( HTML.REPORT %in% output )
	{
		.writeGeneralReport( report, .concat( filename, ".html" ), debug, HTML.REPORT, FALSE, credits=credits, level=level, debugCss=debugCss, debugJavaScript=debugJavaScript );		
	}
	
	if ( HTML.FRAGMENT %in% output )
	{
		warning( "HTML fragments are no longer supported in Nozzle version 0.3.0 or later." );
		#.writeGeneralReport( report, .concat( filename, ".html" ), debug, HTML.FRAGMENT, FALSE, level=level, debugCss=debugCss, debugJavaScript=debugJavaScript );
	}	
}


.writeGeneralReport <- function( report, filename, debug=FALSE, output=HTML.REPORT, append=FALSE, credits=TRUE, level=PROTECTION.PUBLIC, debugCss=NA, debugJavaScript=NA )
{
	# initialize counters for all numbered objects (figures, tables, references, etc.)
	.initNumbers();
							
	report <- .removeProtectedElements( report, level );
	
	# extract result elements (e.g. from paragraphs) and insert into the report tree
	report <- .updateResults( report, report=report )$element; # if there are results directly in the report they will be ignored
	report <- .updateNumberedElements( report );
	report <- .updateReferences( report, report=report );
	
	file <- NA;
	
	if ( append )
	{
		file <- file( filename, "a" );
	}
	else
	{
		file <- file( filename, "w" );
	}
	
	# write file header if in debug mode or if writing standalone report
	if ( output == HTML.REPORT )
	{
		.write( "<!DOCTYPE html>", file );
		
		.write( .tag( "html" ), file );
		
		.write( .tag( "head" ), file );
		
		# write report title
		.write( .tag( "title" ), file );
		.write( report$title, file );
		.write( .tag( "/title" ), file );					
		
		# embedded Google Analytics JavaScript if a Google Analytics id has been provided (should be last before </head>)(
		if ( !is.na( getGoogleAnalyticsId( report ) ) )
		{
			.write( "<script type=\"text/javascript\">", file );
			.write( paste( gsub( .NOZZLE.GOOGLEANALYTICS.VARIABLE, getGoogleAnalyticsId( report ), get( "googleAnalyticsCode", env=.nozzleEnv ) ), collapse="\n" ), file );
			.write( "</script>", file );	
		}		
		
		.write( .tag( "/head" ), file );
		
		.write( .tag( "body" ), file );
		
		# insert mask to make report loading appear more gracefully
		# the mask covers the view and hides possible ugly artefacts that might appear during loading before
		# all divs have been resized to their correct size
		.write( "<div style=\"display: none\" id=\"mask\">&nbsp;</div>", file );
		.write( "<script type=\"text/javascript\">", file );		
		.write( "document.getElementById(\"mask\").style.display = \"block\";", file );
		.write( "</script>", file );			
	}	
	
	.write( .tag( "div", class="report", id=report$reportId ), file );
	
	.write( .comment( "Nozzle Report Id = ", report$reportId ), file );
	.write( .comment( "Nozzle Report Version = ", report$meta$version ), file );

	.write( .comment( "Nozzle Report Create Package = ", getCreatorName( report ) ), file );
	.write( .comment( "Nozzle Report Create Date = ", getCreatorDate( report ) ), file );
	
	report$meta$renderer$name <- .getPackageVersion();
	report$meta$renderer$date <- date();

	.write( .comment( "Nozzle Report Render Package = ", getRendererName( report ) ), file );
	.write( .comment( "Nozzle Report Render Date = ", getRendererDate( report ) ), file );
	
	# add further meta information
	report$meta$reportTitle <- report$title;
	report$meta$reportId <- report$reportId;
	
		
	# insert Nozzle CSS
	if ( debug )
	{
		if ( is.na( debugCss ) )
		{
			.write( .tag( "style media=\"screen\"" ), file );	
			.write( paste( get( "cssCode", env=.nozzleEnv ), collapse="\n" ), file );
			.write( .tag( "/style" ), file );

			.write( .tag( "style media=\"print\"" ), file );	
			.write( paste( get( "cssPrintCode", env=.nozzleEnv ), collapse="\n" ), file );
			.write( .tag( "/style" ), file );
		}
		else
		{
			.write( "<link type=\"text/css\" rel=\"stylesheet\" href=\"", debugCss, "\">", file );
		}
	}
	else
	{
		.write( .tag( "style media=\"screen\"" ), file );	
		.write( paste( get( "cssMinCode", env=.nozzleEnv ), collapse="\n" ), file );	
		.write( .tag( "/style" ), file );
		
		.write( .tag( "style media=\"print\"" ), file );	
		.write( paste( get( "cssPrintMinCode", env=.nozzleEnv ), collapse="\n" ), file );
		.write( .tag( "/style" ), file );		
	}
	
	if ( !is.na( getCustomScreenCss( report ) ) )
	{
		.write( "<link type=\"text/css\" rel=\"stylesheet\" media=\"screen\" href=\"", getCustomScreenCss( report ), "\">", file );
	}
	
	if ( !is.na( getCustomPrintCss( report ) ) )
	{
		.write( "<link type=\"text/css\" rel=\"stylesheet\" media=\"print\" href=\"", getCustomPrintCss( report ), "\">", file );
	}	
	
	# embedded jQuery JavaScript (only in standalone)
	if ( output == HTML.REPORT )
	{
		#.write( "<script type=\"text/javascript\" src=\"https://ajax.googleapis.com/ajax/libs/jquery/1.5.0/jquery.min.js\"></script>", file );
		.write( "<script type=\"text/javascript\">", file );
		.write( paste( get( "jQueryMinCode", env=.nozzleEnv ), collapse="\n" ), file );	
		.write( "</script>", file );	
	}
	
	# embedded TableSorter JavaScript
	.write( "<script type=\"text/javascript\">", file );
	if ( debug )
	{
		.write( paste( get( "jQueryTableSorterMinCode", env=.nozzleEnv ), collapse="\n" ), file );
	}
	else
	{
		.write( paste( get( "jQueryTableSorterMinCode", env=.nozzleEnv ), collapse="\n" ), file );	
	}
	.write( "</script>", file );	
	

	# embedded Nozzle JavaScript
	if ( debug )
	{
		if ( is.na( debugJavaScript ) )
		{
			.write( "<script type=\"text/javascript\">", file );
			.write( paste( get( "javaScriptCode", env=.nozzleEnv ), collapse="\n" ), file );
			.write( "</script>", file );	
		}
		else
		{
			.write( "<script type=\"text/javascript\" src=\"", debugJavaScript ,"\"></script>", file );			
		}
	}
	else
	{
		.write( "<script type=\"text/javascript\">", file );
		.write( paste( get( "javaScriptMinCode", env=.nozzleEnv ), collapse="\n" ), file );	
		.write( "</script>", file );	
	}


	if ( ( output == HTML.REPORT ) || ( append == TRUE ) )
	{
		.write( "<script type=\"text/javascript\">$(document).ready( function() { initNozzle( \"", report$reportId, "\", $ ); });</script>", file );
	}
	else
	{
		# warning!
	}
	
	
	# === REPORT META BEGIN ===	
	
	.write( "<script type=\"text/javascript\">", file );
	.writeJsonElement( "nozzleMeta", report$meta, file );
	.writeJsonElement( "nozzleNavigation", report$navigation, file );
	.write( "</script>", file );	
	
	# === REPORT META END ===	


	# === REPORT CONTENT BEGIN ====

	.write( .tag( "div", class="frame" ), file );	
	.write( .tag( "div", class="main" ), file );
		
	# set global report id if this is being written as a fragment
	# if set to anything else but "", .write will attach this to all ids in HTML tags
	.setGlobalReportId( report$reportId );

	# write top logos
	.writeTopLogos( report, file );
	
	# write title
	.write( .tag( "div", class="title" ), report$title, .tag( "/div" ), file );
	
	# write maintainer info
	if ( !is.na( getMaintainerName( report ) ) )
	{
		.write( .tag( "div", class="maintainer" ), file );
		
		if ( !is.na( getMaintainerEmail( report ) ) )
		{
			.write( "Maintained by ", .tag( "span", class="name" ), asLink( url=.concat( "mailto:", getMaintainerEmail( report ) ), getMaintainerName( report ) ), .tag( "/span" ), file );
		}
		else
		{
			.write( "Maintained by ", .tag( "span", class="name" ), getMaintainerName( report ), .tag( "/span" ), file );
		}

		if ( !is.na( getMaintainerAffiliation( report ) ) )
		{
			.write( " (", .tag( "span", class="affiliation" ), getMaintainerAffiliation( report ), .tag( "/span" ), ")", file );
		}
				
		.write( .tag( "/div" ), file );
	}
	
	# write content
	
	if ( length( report$elements ) > 0 )
	{	
		for ( i in 1:length( report$elements ) )
		{
			.writeElement( report$elements[[i]], file );		
		}
	}
	
	# write copyright info
	if ( !is.na( getCopyrightOwner( report ) ) )
	{
		.write( .tag( "div", class="copyright" ), file );
		.write( .tag( "span", class="notice" ), file );

		.write( "Copyright &copy; ", getCopyrightYear( report ), file );
				
		.write( .tag( "span", class="owner" ), file );
		if ( !is.na( getCopyrightUrl( report ) ) )
		{
			.write( asLink( getCopyrightOwner( report ), url=getCopyrightUrl( report ) ), ".", file );
		}
		else
		{
			.write( getCopyrightOwner( report ), ".", file );
		}
		.write( .tag( "/span" ), file );

		if ( !is.na( getCopyrightStatement( report ) ) )
		{
			.write( getCopyrightStatement( report ), file );
		}

		.write( .tag( "/span" ), file );		
		.write( .tag( "/div" ), file );		
	}
	
	
	# write bottom logos
	.writeBottomLogos( report, file );	


	# write credits (something like "MADE WITH NOZZLE")
	if ( credits )
	{
		.write( .tag( "div", class="credits" ), file );
		.write( .tag( "span", class="notice" ), file );

		.write( asLink( .NOZZLE.CREDITS, url=.NOZZLE.URL ), file );

		.write( .tag( "/span" ), file );		
		.write( .tag( "/div" ), file );		
	}

	
	.write( .tag( "/div" ), file ); # main
	.write( .tag( "/div" ), file ); # frame				
	.write( .tag( "/div" ), file ); # report

	# === REPORT CONTENT END ====

	
	# write file footer if in debug mode or standalone report
	if ( output == HTML.REPORT )
	{	
		.write( .tag( "/body" ), file );
		
		.write( .tag( "/html" ), file );	
	}


	# reset global report id if this is being written as a fragment to avoid conflicts with other reports
	.setGlobalReportId( "" );
	
	close( file );
}


# writes a meta entry into the header into file unless the value is NA with write.na == FALSE
.writeMetaInformation <- function( attribute, value, file, write.na=FALSE )
{
	if ( !is.na( value ) || write.na )
	{
		.write( .tag( "meta", other=.concat( "attr=\"", attribute, "\" value=\"", value, "\"" ) ), file );
		.write( .tag( "/meta" ), file );
	}
}


#===================================================================================================
# functions to write R lists a JSON objects
#===================================================================================================

.writeJsonElement <- function( jsVariable, element, file )
{
	.write( "var ", jsVariable, " = {", file );
	
	.writeJsonList( NA, element, file, indent=3 );
		
	.write( "};", file );		
}


.writeJsonNumeric <- function( name, value, file, indent=0 )
{
	if ( length( value ) == 1 )
	{
		.write( .createIndent( indent ), "\"", name, "\": ", value, file, nobreak=T );
	}
	else
	{
		.write( .createIndent( indent ), "\"", name, "\": [", paste( value, collapse=", " ), "]", file, nobreak=T );	
	}
}


.writeJsonCharacter <- function( name, value, file, indent=0 )
{
	if ( length( value ) == 1 )
	{
		.write( .createIndent( indent ), "\"", name, "\": ", "\"", value, "\"", file, nobreak=T );
	}
	else
	{
		.write( .createIndent( indent ), "\"", name, "\": [", paste( paste( "\"", value, "\"", sep="" ), collapse=", " ), "]", file, nobreak=T );	
	}
}


.writeJsonNull <- function( name, file, indent=0 )
{
	.write( .createIndent( indent ), "\"", name, "\": ", "null", file, nobreak=T );
}


.writeJsonList <- function( name, element, file, indent=0 )
{
	if ( !is.na( name ) )
	{
		.write( .createIndent( indent ), "\"", name, "\": {", file );
	}

	# get names	
	listNames <- names( element );
		
	# iterate over names and write out values
	if ( length( listNames ) > 0 )
	{	
		for ( i in 1:length( listNames ) )
		{
			value <- element[[listNames[i]]];
			
			if ( is.list( value ) )
			{
				.writeJsonList( listNames[i], value, file, indent=indent+3 ); 
			}
			else if ( is.na( value ) )
			{
				.writeJsonNull( listNames[i], file, indent=indent+3 ); 
			}
			else if ( is.numeric( value ) )
			{
				.writeJsonNumeric( listNames[i], value, file, indent=indent+3 );
			}
			else		
			{
				.writeJsonCharacter( listNames[i], value, file, indent=indent+3 );		
			}
			
			if ( i < length( listNames ) )
			{
				.write( ",", file );
			}
			else
			{
				.write( "\n", file, nobreak=T );
			}
		}
	}
	
	# closing braces
	if ( !is.na( name ) )
	{
		.write( .createIndent( indent ), "}", file, nobreak=T );
	}
}

.createIndent <- function( indent, character=" " )
{
	return( paste( rep( character, indent ), collapse="" ) );
}